"""API for the Sheets app"""
import datetime
import json
import logging
import os
import pickle
from collections import namedtuple
from urllib.parse import urljoin

import pygsheets
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google.oauth2.service_account import Credentials as ServiceAccountCredentials
from googleapiclient.discovery import build

from mitol.common.utils import now_in_utc
from mitol.google_sheets.constants import (
    DEFAULT_GOOGLE_EXPIRE_TIMEDELTA,
    GOOGLE_API_FILE_WATCH_KIND,
    GOOGLE_API_NOTIFICATION_TYPE,
    GOOGLE_TOKEN_URI,
    REQUIRED_GOOGLE_API_SCOPES,
)
from mitol.google_sheets.models import GoogleApiAuth
from mitol.google_sheets.utils import format_datetime_for_google_timestamp

log = logging.getLogger(__name__)

DEV_TOKEN_PATH = "localdev/google.token"
FileWatchSpec = namedtuple(
    "FileWatchSpec",
    ["sheet_metadata", "sheet_file_id", "channel_id", "handler_url", "force"],
)


def get_google_creds_from_pickled_token_file(token_file_path):
    """
    Helper method to get valid credentials from a local token file (and refresh as necessary).
    For dev use only.
    """
    with open(token_file_path, "rb") as f:
        creds = pickle.loads(f.read())
    if creds and creds.expired and creds.refresh_token:
        creds.refresh(Request())
        with open(token_file_path, "wb") as token:
            pickle.dump(creds, token)
    if not creds:
        raise ImproperlyConfigured("Local token file credentials are empty")
    if not creds.valid:
        raise ImproperlyConfigured("Local token file credentials are invalid")
    return creds


def get_credentials():
    """
    Gets valid Google API client credentials

    Returns:
        google.oauth2.credentials.Credentials: Credentials to be used by the Google Drive client/pygsheets/etc.

    Raises:
        ImproperlyConfigured: Raised if no credentials have been configured
    """
    if settings.MITOL_GOOGLE_SHEETS_DRIVE_SERVICE_ACCOUNT_CREDS:
        is_sharing_to_service_account = any(
            email
            for email in settings.SHEETS_ADMIN_EMAILS
            if email.endswith(settings.MITOL_GOOGLE_SHEETS_GOOGLE_ACCOUNT_EMAIL_DOMAIN)
        )
        if not is_sharing_to_service_account:
            raise ImproperlyConfigured(
                "If Service Account auth is being used, the SHEETS_ADMIN_EMAILS setting must "
                "include a Service Account email for spreadsheet updates/creation to work. "
                "Add the Service Account email to that setting, or remove the MITOL_GOOGLE_SHEETS_DRIVE_SERVICE_ACCOUNT_CREDS "
                "setting and use a different auth method."
            )
        return ServiceAccountCredentials.from_service_account_info(
            json.loads(settings.MITOL_GOOGLE_SHEETS_DRIVE_SERVICE_ACCOUNT_CREDS),
            scopes=REQUIRED_GOOGLE_API_SCOPES,
        )
    google_api_auth = GoogleApiAuth.objects.order_by("-updated_on").first()
    if google_api_auth:
        creds = Credentials(
            token=google_api_auth.access_token,
            refresh_token=google_api_auth.refresh_token,
            token_uri=GOOGLE_TOKEN_URI,
            client_id=settings.MITOL_GOOGLE_SHEETS_DRIVE_CLIENT_ID,
            client_secret=settings.MITOL_GOOGLE_SHEETS_DRIVE_CLIENT_SECRET,
            scopes=REQUIRED_GOOGLE_API_SCOPES,
        )
        # Proactively refresh if necessary
        needs_refresh = (
            creds.expired
            if creds.expiry
            else google_api_auth.updated_on
            < (now_in_utc() - datetime.timedelta(**DEFAULT_GOOGLE_EXPIRE_TIMEDELTA))
        )
        if needs_refresh:
            log.info("Refreshing GoogleApiAuth credentials...")
            creds.refresh(Request())
            GoogleApiAuth.objects.filter(id=google_api_auth.id).update(
                access_token=creds.token, updated_on=now_in_utc()
            )
        return creds
    # (For local development use only) You can use a locally-created token for auth.
    # This token can be created by following the Google API Python quickstart guide:
    # https://developers.google.com/sheets/api/quickstart/python.
    # A script with more helpful options than the one in that guide can be found here:
    # https://gist.github.com/gsidebo/b87abaafda3e79186c1e5f7f964074ab
    if settings.ENVIRONMENT == "dev":
        token_file_path = os.path.join(settings.BASE_DIR, DEV_TOKEN_PATH)
        if os.path.exists(token_file_path):
            return get_google_creds_from_pickled_token_file(token_file_path)
    raise ImproperlyConfigured("Authorization with Google has not been completed.")


def get_authorized_pygsheets_client():
    """
    Instantiates a pygsheets Client and authorizes it with the proper credentials.

    Returns:
        pygsheets.client.Client: The authorized Client object
    """
    credentials = get_credentials()
    pygsheets_client = pygsheets.authorize(custom_credentials=credentials)
    if settings.MITOL_GOOGLE_SHEETS_DRIVE_SHARED_ID:
        pygsheets_client.drive.enable_team_drive(
            team_drive_id=settings.MITOL_GOOGLE_SHEETS_DRIVE_SHARED_ID
        )
    return pygsheets_client


class ExpandedSheetsClient:
    """
    Helper class that executes some Drive/Sheets API requests that pygsheets doesn't directly support
    """

    def __init__(self, pygsheets_client):
        """
        Args:
            pygsheets_client (pygsheets.client.Client): An authorized pygsheets client
        """
        self.pygsheets_client = pygsheets_client
        self.supports_team_drives = bool(settings.MITOL_GOOGLE_SHEETS_DRIVE_SHARED_ID)

    def get_metadata_for_matching_files(self, query, file_fields="id, name"):
        """
        Fetches metadata for all Drive files that match a given query
        Args:
            query (str): The Drive files query (ref: https://developers.google.com/drive/api/v3/search-files)
            file_fields (str): Comma-separated list of file fields that should be returned in the metadata
                results (ref: https://developers.google.com/drive/api/v3/reference/files#resource)

        Returns:
            list of dict: A dict of metadata for each file that matched the given query
        """
        extra_list_params = {}
        if self.supports_team_drives:
            extra_list_params.update(
                dict(
                    corpora="teamDrive",
                    teamDriveId=settings.MITOL_GOOGLE_SHEETS_DRIVE_SHARED_ID,
                    supportsTeamDrives=True,
                    includeTeamDriveItems=True,
                )
            )
        return self.pygsheets_client.drive.list(
            **extra_list_params, fields="files({})".format(file_fields), q=query
        )

    def update_spreadsheet_properties(self, file_id, property_dict):
        """
        Sets metadata properties on the spreadsheet, which can then be
        included in queries.

        Args:
            file_id (str): The spreadsheet ID
            property_dict (dict): Dict of properties to set

        Returns:
            dict: Google Drive API response to the files.update request
        """
        return (
            self.pygsheets_client.drive.service.files()
            .update(
                fileId=file_id,
                body={"appProperties": property_dict},
                supportsTeamDrives=self.supports_team_drives,
            )
            .execute()
        )

    def get_drive_file_metadata(self, file_id, fields="id, name, modifiedTime"):
        """
        Helper method to fetch metadata for some Drive file.

        Args:
           file_id (str): The file ID
           fields (str): Comma-separated list of file fields that should be returned in the metadata
                results (ref: https://developers.google.com/drive/api/v3/reference/files#resource)

        Returns:
           dict: The file metadata, which includes the specified fields.
        """
        return (
            self.pygsheets_client.drive.service.files()
            .get(
                fileId=file_id,
                fields=fields,
                supportsTeamDrives=self.supports_team_drives,
            )
            .execute()
        )

    def get_sheet_properties(self, file_id):
        """
        Helper method to fetch the dictionary of appProperties for the given spreadsheet by its
        file ID. The Google Sheets API doesn't know about this data. Only the Drive API can access
        it, hence this helper method.

        Args:
            file_id (str): The spreadsheet ID

        Returns:
            dict: appProperties (if any) for the given sheet according to Drive
        """
        result = self.get_drive_file_metadata(file_id=file_id, fields="appProperties")
        if result and "appProperties" in result:
            return result["appProperties"]
        return {}

    def batch_update_sheet_cells(self, sheet_id, request_objects):
        """
        Performs a batch update of targeted cells in a spreadsheet.

        Args:
            sheet_id (str): The spreadsheet id
            request_objects (list of dict): Update request objects
                (docs: https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#Request)

        Returns:
            dict: Google API response to the spreadsheets.values.batchUpdate request
        """
        return (
            self.pygsheets_client.sheet.service.spreadsheets()
            .batchUpdate(spreadsheetId=sheet_id, body={"requests": request_objects})
            .execute()
        )


def build_drive_service(credentials=None):
    """
    Builds the Google API client Drive service for API functionality that cannot be implemented correctly
    with pygsheets.

    Args:
        credentials (google.oauth2.credentials.Credentials or None): Credentials to be used by the
            Google Drive client

    Returns:
        googleapiclient.discovery.Resource: The Drive API service. The methods available on this resource are
            defined dynamically (ref: http://googleapis.github.io/google-api-python-client/docs/dyn/drive_v3.html)
    """
    credentials = credentials or get_credentials()
    return build("drive", "v3", credentials=credentials, cache_discovery=False)


def request_file_watch(
    file_id, channel_id, handler_url, expiration=None, credentials=None
):
    """
    Sends a request to the Google API to watch for changes in a given file. If successful, this
    app will receive requests from Google when changes are made to the file.
    Ref: https://developers.google.com/drive/api/v3/reference/files/watch
    Args:
        file_id (str): The id of the file in Google Drive (can be determined from the URL)
        channel_id (str): Arbitrary string to identify the file watch being set up. This will
            be included in the header of every request Google sends to the app.
        handler_url (str): The URL stub for the xpro endpoint that should be called from Google's end when the file
            changes.
        expiration (datetime.datetime or None): The datetime that this file watch should expire.
            Defaults to 1 hour, and cannot exceed 24 hours.
        credentials (google.oauth2.credentials.Credentials or None): Credentials to be used by the
            Google Drive client
    Returns:
        dict: The Google file watch API response
    """
    drive_service = build_drive_service(credentials=credentials)
    extra_body_params = {}
    if expiration:
        extra_body_params["expiration"] = format_datetime_for_google_timestamp(
            expiration
        )
    return (
        drive_service.files()
        .watch(
            fileId=file_id,
            supportsTeamDrives=True,
            body={
                "id": channel_id,
                "resourceId": file_id,
                "address": urljoin(settings.SITE_BASE_URL, handler_url),
                "payload": True,
                "kind": GOOGLE_API_FILE_WATCH_KIND,
                "type": GOOGLE_API_NOTIFICATION_TYPE,
                **extra_body_params,
            },
        )
        .execute()
    )



"""HTTP views for sheets app"""
import logging
from urllib.parse import urljoin

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.db import transaction
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views.decorators.csrf import csrf_exempt
from google.auth.exceptions import GoogleAuthError
from google_auth_oauthlib.flow import Flow

from mitol.google_sheets.constants import REQUIRED_GOOGLE_API_SCOPES
from mitol.google_sheets.models import GoogleApiAuth
from mitol.google_sheets.utils import generate_google_client_config

log = logging.getLogger(__name__)


@staff_member_required(login_url="login")
def sheets_admin_view(request):
    """Admin view that renders a page that allows a user to begin Google OAuth auth"""
    existing_api_auth = GoogleApiAuth.objects.first()
    successful_action = request.GET.get("success")
    return render(
        request,
        "admin.html",
        {
            "existing_api_auth": existing_api_auth,
            "auth_completed": successful_action == "auth",
        },
    )


@staff_member_required(login_url="login")
def request_google_auth(request):
    """Admin view to begin Google OAuth auth"""
    flow = Flow.from_client_config(
        generate_google_client_config(), scopes=REQUIRED_GOOGLE_API_SCOPES
    )
    flow.redirect_uri = urljoin(
        settings.SITE_BASE_URL, reverse("google-sheets:complete-google-auth")
    )
    authorization_url, state = flow.authorization_url(
        access_type="offline", include_granted_scopes="true", prompt="consent"
    )
    request.session["state"] = state
    request.session["code_verifier"] = flow.code_verifier
    return redirect(authorization_url)


@csrf_exempt
def complete_google_auth(request):
    """Admin view that handles the redirect from Google after completing Google auth"""
    state = request.session.get("state")
    if not state:
        raise GoogleAuthError(
            "Could not complete Google auth - 'state' was not found in the session"
        )
    flow = Flow.from_client_config(
        generate_google_client_config(), scopes=REQUIRED_GOOGLE_API_SCOPES, state=state
    )
    flow.redirect_uri = urljoin(
        settings.SITE_BASE_URL, reverse("google-sheets:complete-google-auth")
    )
    flow.code_verifier = request.session["code_verifier"]
    flow.fetch_token(code=request.GET.get("code"))

    # Store credentials
    credentials = flow.credentials
    with transaction.atomic():
        google_api_auth, _ = GoogleApiAuth.objects.select_for_update().get_or_create()
        google_api_auth.requesting_user = request.user
        google_api_auth.access_token = credentials.token
        google_api_auth.refresh_token = credentials.refresh_token
        google_api_auth.save()

    return redirect(
        "{}?success=auth".format(reverse("google-sheets:sheets-admin-view"))
    )



"""Endpoint URLs for sheets app"""
from django.urls import re_path

from mitol.google_sheets import views

app_name = "google-sheets"

urlpatterns = [
    re_path(r"^sheets/admin/", views.sheets_admin_view, name="sheets-admin-view"),
    re_path(r"^sheets/auth/", views.request_google_auth, name="request-google-auth"),
    re_path(
        r"^sheets/auth-complete/",
        views.complete_google_auth,
        name="complete-google-auth",
    ),
]



"""google sheets settings """
import pytz

from mitol.common.envs import get_string

MITOL_GOOGLE_SHEET_PROCESSOR_APP_NAME = get_string(
    name="MITOL_GOOGLE_SHEET_PROCESSOR_APP_NAME",
    description="Name of the app processing the request",
    required=False,
)
MITOL_GOOGLE_SHEETS_GOOGLE_ACCOUNT_EMAIL_DOMAIN = get_string(
    name="MITOL_GOOGLE_SHEETS_GOOGLE_ACCOUNT_EMAIL_DOMAIN",
    description="Email domain of the google service account",
    default="iam.gserviceaccount.com",
    required=False,
)

MITOL_GOOGLE_SHEETS_DRIVE_SERVICE_ACCOUNT_CREDS = get_string(
    name="MITOL_GOOGLE_SHEETS_DRIVE_SERVICE_ACCOUNT_CREDS",
    default=None,
    description="The contents of the Service Account credentials JSON to use for Google API auth",
)
MITOL_GOOGLE_SHEETS_DRIVE_CLIENT_ID = get_string(
    name="MITOL_GOOGLE_SHEETS_DRIVE_CLIENT_ID",
    default=None,
    description="Client ID from Google API credentials",
)
MITOL_GOOGLE_SHEETS_DRIVE_CLIENT_SECRET = get_string(
    name="MITOL_GOOGLE_SHEETS_DRIVE_CLIENT_SECRET",
    default=None,
    description="Client secret from Google API credentials",
)
MITOL_GOOGLE_SHEETS_DRIVE_API_PROJECT_ID = get_string(
    name="MITOL_GOOGLE_SHEETS_DRIVE_API_PROJECT_ID",
    default=None,
    description="ID for the Google API project where the credentials were created",
)
MITOL_GOOGLE_SHEETS_DRIVE_SHARED_ID = get_string(
    name="MITOL_GOOGLE_SHEETS_DRIVE_SHARED_ID",
    default=None,
    description="ID of the Shared Drive (a.k.a. Team Drive). This is equal to the top-level folder ID.",
)
MITOL_GOOGLE_SHEETS_ENROLLMENT_CHANGE_SHEET_ID = get_string(
    name="MITOL_GOOGLE_SHEETS_ENROLLMENT_CHANGE_SHEET_ID",
    default=None,
    description=(
        "ID of the Google Sheet that contains the enrollment change request worksheets (refunds, transfers, etc)"
    ),
)
MITOL_GOOGLE_SHEETS_DATE_FORMAT = get_string(
    name="MITOL_GOOGLE_SHEETS_DATE_FORMAT",
    default="%m/%d/%Y",
    description="Python strptime format for date columns (no time) in enrollment management spreadsheets",
)
_sheets_date_timezone = get_string(
    name="MITOL_GOOGLE_SHEETS_DATE_TIMEZONE",
    default="UTC",
    description=(
        "The name of the timezone that should be assumed for date/time values in spreadsheets. "
        "Choose from a value in the TZ database (https://en.wikipedia.org/wiki/List_of_tz_database_time_zones)."
    ),
)
MITOL_GOOGLE_SHEETS_DATE_TIMEZONE = pytz.timezone(_sheets_date_timezone)



"""Sheets app models"""
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import DateTimeField, Model, PositiveSmallIntegerField

from mitol.common.models import SingletonModel, TimestampedModel


class GoogleApiAuth(TimestampedModel, SingletonModel):
    """Model that stores OAuth credentials to be used to authenticate with the Google API"""

    requesting_user = models.OneToOneField(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True
    )
    access_token = models.CharField(max_length=2048)
    refresh_token = models.CharField(null=True, max_length=512)


class GoogleFileWatch(TimestampedModel):
    """
    Model that represents a file watch/push notification/webhook that was set up via the Google API for
    some Google Drive file
    """

    file_id = models.CharField(max_length=100, db_index=True, null=False)
    channel_id = models.CharField(max_length=100, db_index=True, null=False)
    version = models.IntegerField(db_index=True, null=True, blank=True)
    activation_date = models.DateTimeField(null=False)
    expiration_date = models.DateTimeField(null=False)
    last_request_received = models.DateTimeField(null=True, blank=True)

    class Meta:
        unique_together = ("file_id", "version")

    def save(
        self, force_insert=False, force_update=False, using=None, update_fields=None
    ):
        if (
            force_insert
            and self._meta.model.objects.filter(file_id=self.file_id).count() > 0
        ):
            raise ValidationError(
                "Only one {} object should exist for each unique file_id (file_id provided: {}). "
                "Update the existing object instead of creating a new one.".format(
                    self.__class__.__name__, self.file_id
                )
            )
        return super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )

    def __str__(self):
        return "GoogleFileWatch: id={}, channel_id={}, file_id={}, expires={}".format(
            self.id, self.channel_id, self.file_id, self.expiration_date.isoformat()
        )


class FileWatchRenewalAttempt(Model):
    """
    Tracks attempts to renew a Google file watch. Used for debugging flaky endpoint.
    """

    sheet_type = models.CharField(max_length=30, db_index=True, null=False)
    sheet_file_id = models.CharField(max_length=100, db_index=True, null=False)
    date_attempted = DateTimeField(auto_now_add=True)
    result = models.CharField(max_length=300, null=True, blank=True)
    result_status_code = PositiveSmallIntegerField(null=True, blank=True)


class GoogleSheetsRequestModel(TimestampedModel):
    """Model that represents a request to change an enrollment"""

    form_response_id = models.IntegerField(db_index=True, unique=True, null=False)
    date_completed = models.DateTimeField(null=True, blank=True)
    raw_data = models.CharField(max_length=300, null=True, blank=True)

    class Meta:
        abstract = True



# Generated by Django 3.2.13 on 2022-06-17 14:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("google_sheets", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="filewatchrenewalattempt",
            name="sheet_type",
            field=models.CharField(db_index=True, max_length=30),
        ),
    ]



"""Sheets app constants"""

REQUIRED_GOOGLE_API_SCOPES = [
    "https://www.googleapis.com/auth/spreadsheets",
    "https://www.googleapis.com/auth/drive",
]
DEFAULT_GOOGLE_EXPIRE_TIMEDELTA = dict(minutes=60)

SHEETS_VALUE_REQUEST_PAGE_SIZE = 50

# The index of the first row of a spreadsheet according to Google
GOOGLE_SHEET_FIRST_ROW = 1
GOOGLE_AUTH_URI = "https://accounts.google.com/o/oauth2/auth"
GOOGLE_TOKEN_URI = "https://oauth2.googleapis.com/token"
GOOGLE_AUTH_PROVIDER_X509_CERT_URL = "https://www.googleapis.com/oauth2/v1/certs"
GOOGLE_DATE_TIME_FORMAT = "DATE_TIME"
GOOGLE_API_NOTIFICATION_TYPE = "webhook"
GOOGLE_API_FILE_WATCH_KIND = "api#channel"
GOOGLE_API_TRUE_VAL = "TRUE"



"""API with general functionality for all enrollment change spreadsheets"""
import json
import logging
import operator as op

from django.conf import settings
from django.db import transaction
from django.utils.functional import cached_property

from mitol.common.utils.collections import group_into_dict
from mitol.google_sheets.api import get_authorized_pygsheets_client
from mitol.google_sheets.constants import GOOGLE_SHEET_FIRST_ROW
from mitol.google_sheets.utils import (
    ResultType,
    RowResult,
    format_datetime_for_sheet_formula,
    get_data_rows,
    get_data_rows_after_start,
)

log = logging.getLogger(__name__)


class SheetHandler:
    """
    Base class for managing the processing of a spreadsheet which contains requests for work to be done
    (creating enrollment codes, changing enrollments, etc.).
    """

    pygsheets_client = None
    spreadsheet = None
    sheet_metadata = None

    @cached_property
    def worksheet(self):
        """
        Returns the correct Worksheet object for this spreadsheet

        Returns:
             pygsheets.worksheet.Worksheet: The Worksheet object
        """
        # By default, the first worksheet of the spreadsheet should be used
        return self.spreadsheet.sheet1

    def get_enumerated_rows(self):
        """
        Yields enumerated data rows of a spreadsheet (excluding header row(s))

        Yields:
            Tuple[int, List[str]]: Row index (according to the Google Sheet, NOT zero-indexed) paired with the list
                of strings representing the data in each column of the row
        """
        yield from enumerate(
            get_data_rows(self.worksheet, include_trailing_empty=False),
            start=GOOGLE_SHEET_FIRST_ROW + 1,
        )

    def update_completed_rows(self, success_row_results):
        """
        Updates rows in the spreadsheet that were successfully processed.

        Args:
            success_row_results (Iterable[RowResult]): Objects representing the results of processing a row
        """
        raise NotImplementedError

    def update_row_errors(self, failed_row_results):
        """
        Updates rows in the spreadsheet that failed during processing.

        Args:
            failed_row_results (Iterable[RowResult]): Objects representing the results of processing a row
        """
        for row_result in failed_row_results:
            self.worksheet.update_value(
                "{}{}".format(
                    self.sheet_metadata.ERROR_COL_LETTER, row_result.row_index
                ),
                row_result.message,
            )

    def update_sheet_from_results(self, grouped_row_results):
        """
        Updates the relevant spreadsheet cells based on the row results and logs any necessary messages.

        Args:
            grouped_row_results (Dict[str, Iterable[RowResult]]): Objects representing the results of processing rows
                grouped by result type (success, failed, etc.)
        """
        processed_row_results = grouped_row_results.get(ResultType.PROCESSED, [])
        if processed_row_results:
            self.update_completed_rows(processed_row_results)
        failed_row_results = grouped_row_results.get(ResultType.FAILED, [])
        if failed_row_results:
            self.update_row_errors(failed_row_results)
        out_of_sync_row_results = grouped_row_results.get(ResultType.OUT_OF_SYNC, [])
        if out_of_sync_row_results:
            log.warning(
                "Rows found without a completed date, but local records indicate that they were completed: %s",
                [row_result.row_index for row_result in out_of_sync_row_results],
            )
            self.update_completed_rows(out_of_sync_row_results)
        ignored_row_results = grouped_row_results.get(ResultType.IGNORED, [])
        if ignored_row_results:
            log.warning(
                "Ignored rows in %s (%s): %s",
                self.sheet_metadata.sheet_name,
                self.sheet_metadata.worksheet_name,
                [row_result.row_index for row_result in ignored_row_results],
            )

    def post_process_results(self, grouped_row_results):
        """
        Helper method for executing any logic after the sheet processing results have been handled

        Args:
            grouped_row_results (Dict[str, Iterable[RowResult]]): Objects representing the results of processing rows
                grouped by result type (success, failed, etc.)
        """
        return grouped_row_results

    def get_or_create_request(self, row_data):
        """
        Ensures that an object exists in the database that represents this spreadsheet row, and
        that it reflects the correct state based on the raw data.

        Args:
            row_data (List[str]): Raw data from a row in the spreadsheet

        Returns:
            Tuple[Type(GoogleSheetsRequestModel), bool, bool]: A tuple containing an object representing the
                request, a flag that indicates whether or not it was newly created, and a flag that indicates
                whether or not it was updated.
        """
        raise NotImplementedError

    @staticmethod
    def validate_sheet(enumerated_rows):
        """
        Checks the request sheet data for any data issues beyond the scope of a single row (i.e.: any row data
        that is invalid because of the data in other rows in the sheet), and returns the valid rows paired
        with row results for the invalid rows. By default, all rows are considered valid.

        Args:
            enumerated_rows (Iterable[Tuple[int, List[str]]]): Row indices paired with a list of strings
                representing the data in each row

        Returns:
            Tuple[ Iterable[Tuple[int, List[str]]], List[RowResult] ]: Enumerated data rows with invalidated rows
                filtered out, paired with objects representing the rows that failed validation.
        """
        return enumerated_rows, []

    def filter_ignored_rows(self, enumerated_rows):
        """
        Takes an iterable of enumerated rows, and returns an iterable of those rows without the ones that should be
        ignored. By default, no rows are filtered.

        Args:
            enumerated_rows (Iterable[Tuple[int, List[str]]]): Row indices paired with a list of strings
                representing the data in each row

        Returns:
            Iterable[Tuple[int, List[str]]]: Iterable of data rows without the ones that should be ignored.
        """
        return enumerated_rows

    def process_row(self, row_index, row_data):
        """
        Ensures that the given spreadsheet row is correctly represented in the database,
        attempts to parse it, performs the necessary logic to service the request (e.g.: create enrollment codes),
        and returns the result of processing the row.

        Args:
            row_index (int): The row index according to the spreadsheet
            row_data (List[str]): The raw data of the given spreadsheet row

        Returns:
            Optional[RowResult]: An object representing the results of processing the row, or None if
                nothing needs to be done with this row.
        """
        raise NotImplementedError

    def process_sheet(self, limit_row_index=None):
        """
        Ensures that all non-legacy rows in the spreadsheet are correctly represented in the database,
        changes enrollments if appropriate, updates the spreadsheet to reflect any changes
        made, and returns a summary of those changes.

        Returns:
            dict: A summary of the changes made while processing the enrollment change request sheet
        """
        if limit_row_index is None:
            enumerated_rows = self.get_enumerated_rows()
        else:
            enumerated_rows = [
                (limit_row_index, self.worksheet.get_row(limit_row_index))
            ]
        filtered_rows = self.filter_ignored_rows(enumerated_rows)
        valid_enumerated_rows, row_results = self.validate_sheet(filtered_rows)
        for row_index, row_data in valid_enumerated_rows:
            row_result = None
            try:
                row_result = self.process_row(row_index, row_data)
            except Exception as exc:
                log.exception("Error processing row from google sheets")
                row_result = RowResult(
                    row_index=row_index,
                    row_db_record=None,
                    row_object=None,
                    result_type=ResultType.FAILED,
                    message="Error: {}".format(str(exc)),
                )
            finally:
                if row_result:
                    row_results.append(row_result)
        if not row_results:
            return {}
        grouped_row_results = group_into_dict(
            row_results, key_fn=op.attrgetter("result_type")
        )
        self.update_sheet_from_results(grouped_row_results)
        self.post_process_results(grouped_row_results)
        return {
            result_type.value: [row_result.row_index for row_result in row_results]
            for result_type, row_results in grouped_row_results.items()
        }


class GoogleSheetsChangeRequestHandler(SheetHandler):
    """
    Base class for managing the processing of enrollment change requests from a spreadsheet
    """

    def __init__(
        self, spreadsheet_id, worksheet_id, start_row, sheet_metadata, request_model_cls
    ):
        """

        Args:
            spreadsheet_id (int):
            worksheet_id (int):
            start_row (int):
            sheet_metadata (Type(SheetConfig)):
            request_model_cls (Type(GoogleSheetsRequestModel)):
        """
        self.pygsheets_client = get_authorized_pygsheets_client()
        self.spreadsheet = self.pygsheets_client.open_by_key(spreadsheet_id)
        self.worksheet_id = worksheet_id
        self.start_row = start_row
        self.sheet_metadata = sheet_metadata
        self.request_model_cls = request_model_cls

    @cached_property
    def worksheet(self):
        return self.spreadsheet.worksheet("id", value=self.worksheet_id)

    def get_enumerated_rows(self):
        # Only yield rows in the spreadsheet that come after the legacy rows
        # (i.e.: the rows of data that were manually entered before we started automating this process)
        return enumerate(
            get_data_rows_after_start(
                self.worksheet,
                start_row=self.start_row,
                start_col=1,
                end_col=self.sheet_metadata.num_columns,
            ),
            start=self.start_row,
        )

    def update_completed_rows(self, success_row_results):
        for row_result in success_row_results:
            self.worksheet.update_values(
                crange="{processor_col}{row_index}:{error_col}{row_index}".format(
                    processor_col=self.sheet_metadata.PROCESSOR_COL_LETTER,
                    error_col=self.sheet_metadata.ERROR_COL_LETTER,
                    row_index=row_result.row_index,
                ),
                values=[
                    [
                        settings.MITOL_GOOGLE_SHEET_PROCESSOR_APP_NAME,
                        format_datetime_for_sheet_formula(
                            row_result.row_db_record.date_completed.astimezone(
                                settings.MITOL_GOOGLE_SHEETS_DATE_TIMEZONE
                            )
                        ),
                        "",
                    ]
                ],
            )

    def get_or_create_request(self, row_data):
        form_response_id = int(
            row_data[self.sheet_metadata.FORM_RESPONSE_ID_COL].strip()
        )
        user_input_json = json.dumps(
            self.sheet_metadata.get_form_input_columns(row_data)
        )
        with transaction.atomic():
            (
                enroll_change_request,
                created,
            ) = self.request_model_cls.objects.select_for_update().get_or_create(
                form_response_id=form_response_id,
                defaults=dict(raw_data=user_input_json),
            )
            raw_data_changed = enroll_change_request.raw_data != user_input_json
            if raw_data_changed:
                enroll_change_request.raw_data = user_input_json
                enroll_change_request.save()
        return enroll_change_request, created, raw_data_changed

    def process_row(self, row_index, row_data):
        raise NotImplementedError



"""Sheets app exceptions"""


class SheetValidationException(Exception):
    """
    General exception for failures during the validation of Sheet data
    """


class SheetUpdateException(Exception):
    """
    General exception for failures while attempting to update a Sheet via API
    """


class SheetRowParsingException(Exception):
    """
    General exception for failures while attempting to parse the data in a Sheet row
    """


class SheetOutOfSyncException(Exception):
    """
    General exception for situations where the data in a spreadsheet does not reflect the state of the database
    """

    def __init__(self, coupon_gen_request, coupon_req_row, msg=None):
        self.coupon_gen_request = coupon_gen_request
        self.coupon_req_row = coupon_req_row
        super().__init__(msg)


class InvalidSheetProductException(Exception):
    """
    Exception for an invalid product entered into the coupon request spreadsheet
    """



"""Sheets app util functions"""
import datetime
import email.utils
from collections import namedtuple
from enum import Enum
from urllib.parse import quote_plus, urljoin

import pytz
from django.conf import settings
from django.urls import reverse

from mitol.common.utils.base import base_register_subclasses_factory
from mitol.google_sheets.constants import (
    GOOGLE_AUTH_PROVIDER_X509_CERT_URL,
    GOOGLE_AUTH_URI,
    GOOGLE_TOKEN_URI,
    SHEETS_VALUE_REQUEST_PAGE_SIZE,
)


def generate_google_client_config():
    """Helper method to generate Google client config based on app settings"""
    return {
        "web": {
            "client_id": settings.MITOL_GOOGLE_SHEETS_DRIVE_CLIENT_ID,
            "client_secret": settings.MITOL_GOOGLE_SHEETS_DRIVE_CLIENT_SECRET,
            "project_id": settings.MITOL_GOOGLE_SHEETS_DRIVE_API_PROJECT_ID,
            "redirect_uris": [
                urljoin(
                    settings.SITE_BASE_URL,
                    reverse("google-sheets:complete-google-auth"),
                )
            ],
            "auth_uri": GOOGLE_AUTH_URI,
            "token_uri": GOOGLE_TOKEN_URI,
            "auth_provider_x509_cert_url": GOOGLE_AUTH_PROVIDER_X509_CERT_URL,
        }
    }


def get_column_letter(column_index):
    """
    Returns the spreadsheet column letter that corresponds to a given index (e.g.: 0 -> 'A', 3 -> 'D')

    Args:
        column_index (int):

    Returns:
        str: The column index expressed as a letter
    """
    if column_index > 25:
        raise ValueError("Cannot generate a column letter past 'Z'")
    uppercase_a_ord = ord("A")
    return chr(column_index + uppercase_a_ord)


class SheetConfigMixin:
    """Metadata for a type of Google Sheet that this app interacts with"""

    sheet_type = None
    sheet_name = None
    worksheet_type = None
    worksheet_name = None
    first_data_row = None
    non_input_column_indices = set()
    num_columns = 0

    @property
    def form_input_column_indices(self):
        """
        Returns indices of columns that contain data entered by a user in a form

        Returns:
            set: Indices of columns that contain data entered by a user in a form
        """
        return set(range(self.num_columns)).difference(self.non_input_column_indices)

    def handler_url_stub(self, file_id=None):
        """
        Returns the URL that Google should send requests to when a change is made to a watched
        spreadsheet.

        Args:
            file_id (str): (Optional) The id of the spreadsheet as it appears in the spreadsheet's URL.
                If the spreadsheet being watched is a singleton, this isn't necessary.

        Returns:
            str: The URL that Google will send file watch requests to
        """
        params = dict(sheet=quote_plus(self.sheet_type))
        if file_id:
            params["fileId"] = file_id
        param_str = "&".join([f"{k}={v}" for k, v in params.items()])
        return "{}?{}".format(reverse("handle-watched-sheet-update"), param_str)

    def get_form_input_columns(self, row_data):
        """
        Returns a list of column values for columns that contain data entered by a user in a form
        (i.e.: no auto-generated values, or values entered by this app)

        Args:
            row_data (iterable): Cell values from a row in a sheet

        Returns:
            list: Values for columns that contain data entered by a user in a form
        """
        return [
            col for i, col in enumerate(row_data) if i in self.form_input_column_indices
        ]


class SingletonSheetConfigMixin(SheetConfigMixin):
    """
    Metadata for a type of Google Sheet that this app interacts with, and of which only one should exist
    """

    sheet_file_id = None


SingletonSheetConfig = base_register_subclasses_factory(SingletonSheetConfigMixin)


class ResultType(Enum):
    """Enum of possible row results"""

    IGNORED = "ignored"
    FAILED = "failed"
    OUT_OF_SYNC = "out_of_sync"
    PROCESSED = "processed"

    def __lt__(self, other):
        return self.value < other.value


RowResult = namedtuple(
    "RowResult", ["row_index", "row_db_record", "row_object", "message", "result_type"]
)
ProcessedRequest = namedtuple(
    "ProcessedRequest", ["row_index", "coupon_req_row", "request_id", "date_processed"]
)
FailedRequest = namedtuple(
    "FailedRequest", ["row_index", "exception", "sheet_error_text"]
)
IgnoredRequest = namedtuple("IgnoredRequest", ["row_index", "coupon_req_row", "reason"])
AssignmentRowUpdate = namedtuple(
    "AssignmentRowUpdate", ["row_index", "status", "status_date", "alternate_email"]
)


def get_data_rows(worksheet, include_trailing_empty=False):
    """
    Yields the data rows of a spreadsheet that has a header row

    Args:
        worksheet (pygsheets.worksheet.Worksheet): Worksheet object
        include_trailing_empty (bool): Whether to include empty trailing cells/values after last non-zero value

    Yields:
        list of str: List of cell values in a given row
    """
    row_iter = iter(
        worksheet.get_all_values(
            # These param names are a typo in the pygsheets library
            include_tailing_empty=include_trailing_empty,
            include_tailing_empty_rows=False,
        )
    )
    try:
        # Skip header row
        next(row_iter)
    except StopIteration:
        return
    yield from row_iter


def get_data_rows_after_start(
    worksheet,
    start_row,
    start_col,
    end_col,
    page_size=SHEETS_VALUE_REQUEST_PAGE_SIZE,
    **kwargs,
):
    """
    Yields the data rows of a spreadsheet starting with a given row and spanning a given column range
    until empty rows are encountered.

    Args:
        worksheet (pygsheets.worksheet.Worksheet): Worksheet object
        start_row (int): Zero-based index of the first row for which you want data returned
        start_col (int): Zero-based index of the start of the column range
        end_col (int):  Zero-based index of the end of the column range
        page_size (int): The number of rows to fetch per individual API request
        kwargs (dict): Option params to pass along to pygsheets.worksheet.Worksheet.get_values

    Yields:
        list of str: List of cell values in a given row
    """
    request_count = 0
    values = []
    while request_count == 0 or (values and len(values) == page_size):
        end_row = start_row + page_size - 1
        values = worksheet.get_values(
            start=(start_row, start_col),
            end=(end_row, end_col),
            include_tailing_empty=True,
            include_tailing_empty_rows=False,
            returnas="matrix",
            **kwargs,
        )
        request_count += 1
        yield from values
        start_row = end_row + 1


def spreadsheet_repr(spreadsheet=None, spreadsheet_metadata=None):
    """
    Returns a simple string representation of a Spreadsheet object

    Args:
        spreadsheet (pygsheets.spreadsheet.Spreadsheet or None):
        spreadsheet_metadata (dict or None): A dict of spreadsheet metadata

    Returns:
        str: String representation of the spreadsheet
    """
    if spreadsheet:
        sheet_id, title = spreadsheet.id, spreadsheet.title
    elif spreadsheet_metadata:
        sheet_id, title = spreadsheet_metadata["id"], spreadsheet_metadata["name"]
    else:
        sheet_id, title = None, None
    if not sheet_id or not title:
        raise ValueError("Invalid spreadsheet/metadata provided")
    return "'{}', id: {}".format(title, sheet_id)


def clean_sheet_value(value):
    """
    Takes a spreadsheet cell value and returns a cleaned version

    Args:
        value (str): A raw spreadsheet cell value

    Returns:
        str or None: A string with whitespace stripped, or None if the resulting value was an empty string
    """
    stripped = value.strip()
    return None if stripped == "" else stripped


def format_datetime_for_google_api(dt):
    """
    String-ifies a datetime value in the format expected by Google APIs

    Args:
        dt (datetime.datetime):

    Returns:
        str: The datetime formatted for use in a Google API request
    """
    return dt.isoformat()


def format_datetime_for_google_timestamp(dt):
    """
    Formats a datetime for use in a Google API request that expects a timestamp
    (e.g.: file watch expiration – https://developers.google.com/drive/api/v3/reference/files/watch#request-body)

    Args:
        dt (datetime.datetime):

    Returns:
        int: The datetime formatted as a timestamp for use in a Google API request
    """
    # Google expects the timestamp to be in milliseconds, not seconds, hence the '* 1000'
    return int(dt.timestamp() * 1000)


def format_datetime_for_mailgun(dt):
    """
    String-ifies a datetime value in the format expected by the Mailgun API

    Args:
        dt (datetime.datetime):

    Returns:
        str: The datetime formatted for use in a Mailgun API request
    """
    return email.utils.format_datetime(dt)


def format_datetime_for_sheet_formula(dt):
    """
    String-ifies a datetime value in a format that will result in a valid date entry in a Google Sheets cell

    Args:
        dt (datetime.datetime):

    Returns:
        str: The datetime formatted for a Google Sheets cell value
    """
    return f"=DATE({dt.year},{dt.month},{dt.day}) + TIME({dt.hour},{dt.minute},{dt.second})"


def _parse_sheet_date_str(date_str, date_format):
    """
    Parses a string that represents a date/datetime and returns the UTC datetime (or None)

    Args:
        date_str (str): The date/datetime string
        date_format (str): The strptime-compatible format string that the string is expected to match

    Returns:
        datetime.datetime or None: The parsed datetime (in UTC) or None
    """
    if not date_str:
        return None
    dt = datetime.datetime.strptime(date_str, date_format).astimezone(
        settings.MITOL_GOOGLE_SHEETS_DATE_TIMEZONE
    )
    return (
        dt
        if settings.MITOL_GOOGLE_SHEETS_DATE_TIMEZONE == pytz.UTC
        else dt.astimezone(pytz.UTC)
    )


def parse_sheet_datetime_str(datetime_str):
    """
    Parses a string that represents a datetime and returns the UTC datetime (or None)

    Args:
        datetime_str (str): The datetime string

    Returns:
        datetime.datetime or None: The parsed datetime (in UTC) or None
    """
    return _parse_sheet_date_str(datetime_str, settings.SHEETS_DATE_FORMAT)


def parse_sheet_date_only_str(date_str):
    """
    Parses a string that represents a date and returns the UTC datetime (or None)

    Args:
        date_str (str): The datetime string

    Returns:
        datetime.datetime or None: The parsed datetime (in UTC) or None
    """
    return _parse_sheet_date_str(date_str, settings.MITOL_GOOGLE_SHEETS_DATE_FORMAT)


def google_timestamp_to_datetime(google_timestamp):
    """
    Parses a timestamp value from a Google API response as a normal datetime (UTC)

    Args:
        google_timestamp (str or int): A timestamp value from a Google API response

    Returns:
        datetime.datetime: The parsed timestamp with UTC timezone
    """
    # Google timestamps are expressed in milliseconds, hence the '/ 1000'
    timestamp_in_seconds = int(google_timestamp) / 1000
    return datetime.datetime.fromtimestamp(timestamp_in_seconds, pytz.UTC)


def google_date_string_to_datetime(google_date_str):
    """
    Parses a datetime string value from a Google API response as a normal datetime (UTC)

    Args:
        google_date_str (str): A datetime string value from a Google API response

    Returns:
        datetime.datetime: The parsed timestamp with UTC timezone
    """
    return datetime.datetime.strptime(
        google_date_str, "%Y-%m-%dT%H:%M:%S.%fZ"
    ).astimezone(pytz.UTC)


def mailgun_timestamp_to_datetime(timestamp):
    """
    Parses a timestamp value from a Mailgun API response as a datetime

    Args:
        timestamp (float): A timestamp value from a Mailgun API response

    Returns:
        datetime.datetime: The parsed timestamp
    """
    return datetime.datetime.fromtimestamp(timestamp, pytz.UTC)


def build_multi_cell_update_request_body(
    row_index, column_index, values, worksheet_id=0
):
    """
    Builds a dict for use in the body of a Google Sheets API batch update request

    Args:
        row_index (int): The index of the cell row that should be updated (starting with 0)
        column_index (int): The index of the first cell column that should be updated (starting with 0)
        values (list of dict): The updates to be performed
        worksheet_id (int):

    Returns:
        dict: A single update request object for use in a Google Sheets API batch update request
    """
    return {
        "updateCells": {
            "range": {
                "sheetId": worksheet_id,
                "startRowIndex": row_index,
                "endRowIndex": row_index + 1,
                "startColumnIndex": column_index,
                "endColumnIndex": column_index + len(values),
            },
            "rows": [{"values": values}],
            "fields": "*",
        }
    }


def build_protected_range_request_body(
    start_row_index,
    num_rows,
    start_col_index,
    num_cols,
    worksheet_id=0,
    warning_only=False,
    description=None,
):
    """
    Builds a request body that will be sent to the Google Sheets API to create a protected range on a spreadsheet.

    Args:
        start_row_index (int): The zero-based index of the row of the range that will be protected
        num_rows (int): The number of rows that this range will span
        start_col_index (int): The zero-based index of the column of the range that will be protected
        num_cols (int): The number of columns that this range will span
        worksheet_id (int): The worksheet id in the given spreadsheet (the first worksheet id is always 0)
        warning_only (bool): If True, the range will be editable, but will display a warning/confirmation dialog
            before edits are accepted
        description (str or None): An optional description for the protected range

    Returns:
        dict: A request body that will be sent to the Google Sheets API to create a protected range
    """
    extra_params = {} if description is None else {"description": description}
    return {
        "addProtectedRange": {
            "protectedRange": {
                "range": {
                    "sheetId": worksheet_id,
                    "startRowIndex": start_row_index,
                    "endRowIndex": start_row_index + num_rows,
                    "startColumnIndex": start_col_index,
                    "endColumnIndex": start_col_index + num_cols,
                },
                "warningOnly": warning_only,
                **extra_params,
            }
        }
    }


def build_drive_file_email_share_request(file_id, email_to_share):
    """
    Builds the body of a Drive file share request

    Args:
        file_id (str): The file id of the Drive file being shared
        email_to_share (str): The email of the user to whom the file will be shared

    Returns:
        dict: A dictionary of parameters for the body of a share request
    """
    added_kwargs = (
        {"sendNotificationEmail": False}
        if email_to_share.endswith(
            settings.MITOL_GOOGLE_SHEETS_GOOGLE_ACCOUNT_EMAIL_DOMAIN
        )
        else {}
    )
    return dict(
        fileId=file_id,
        body={"type": "user", "role": "writer", "emailAddress": email_to_share},
        fields="id",
        supportsTeamDrives=True,
        **added_kwargs,
    )



"""
Admin site bindings for sheets models
"""

from django.contrib import admin
from django.core.exceptions import ValidationError

from mitol.common.admin import SingletonModelAdmin
from mitol.google_sheets import models


class GoogleApiAuthAdmin(SingletonModelAdmin):
    """Admin for GoogleApiAuth"""

    model = models.GoogleApiAuth
    list_display = ("id", "requesting_user")


class GoogleFileWatchAdmin(admin.ModelAdmin):
    """Admin for GoogleFileWatch"""

    model = models.GoogleFileWatch
    list_display = (
        "id",
        "file_id",
        "channel_id",
        "activation_date",
        "expiration_date",
        "last_request_received",
    )
    ordering = ["-expiration_date"]

    def save_form(self, request, form, change):
        if not change:
            file_id = form.cleaned_data["file_id"]
            if self.model.objects.filter(file_id=file_id).exists():
                raise ValidationError(
                    f"Only one GoogleFileWatch object should exist for each unique file_id (file_id provided: {file_id}). "
                    "Update the existing object instead of creating a new one."
                )
        return super().save_form(request, form, change)


class FileWatchRenewalAttemptAdmin(admin.ModelAdmin):
    """Admin for FileWatchRenewalAttempt"""

    model = models.FileWatchRenewalAttempt
    list_display = (
        "id",
        "sheet_type",
        "sheet_file_id",
        "date_attempted",
        "result",
        "result_status_code",
    )
    search_fields = ("sheet_file_id", "result")
    list_filter = ("sheet_type", "result_status_code")
    readonly_fields = ("date_attempted",)
    ordering = ("-date_attempted",)


admin.site.register(models.GoogleApiAuth, GoogleApiAuthAdmin)
admin.site.register(models.GoogleFileWatch, GoogleFileWatchAdmin)
admin.site.register(models.FileWatchRenewalAttempt, FileWatchRenewalAttemptAdmin)



"""Factories for sheets app"""
from factory import Faker, SubFactory
from factory.django import DjangoModelFactory

from mitol.common.factories import UserFactory
from mitol.google_sheets import models


class GoogleApiAuthFactory(DjangoModelFactory):
    requesting_user = SubFactory(UserFactory)
    access_token = Faker("pystr", max_chars=30)
    refresh_token = Faker("pystr", max_chars=30)

    class Meta:
        model = models.GoogleApiAuth



# Generated by Django 3.2.13 on 2022-06-14 12:36

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="FileWatchRenewalAttempt",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sheet_type",
                    models.CharField(
                        choices=[
                            ("enrollrequest", "enrollrequest"),
                            ("enrollassign", "enrollassign"),
                            ("enrollchange", "enrollchange"),
                        ],
                        db_index=True,
                        max_length=30,
                    ),
                ),
                ("sheet_file_id", models.CharField(db_index=True, max_length=100)),
                ("date_attempted", models.DateTimeField(auto_now_add=True)),
                ("result", models.CharField(blank=True, max_length=300, null=True)),
                (
                    "result_status_code",
                    models.PositiveSmallIntegerField(blank=True, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="GoogleFileWatch",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_on", models.DateTimeField(auto_now_add=True)),
                ("updated_on", models.DateTimeField(auto_now=True)),
                ("file_id", models.CharField(db_index=True, max_length=100)),
                ("channel_id", models.CharField(db_index=True, max_length=100)),
                ("version", models.IntegerField(blank=True, db_index=True, null=True)),
                ("activation_date", models.DateTimeField()),
                ("expiration_date", models.DateTimeField()),
                ("last_request_received", models.DateTimeField(blank=True, null=True)),
            ],
            options={
                "unique_together": {("file_id", "version")},
            },
        ),
        migrations.CreateModel(
            name="GoogleApiAuth",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_on", models.DateTimeField(auto_now_add=True)),
                ("updated_on", models.DateTimeField(auto_now=True)),
                ("access_token", models.CharField(max_length=2048)),
                ("refresh_token", models.CharField(max_length=512, null=True)),
                (
                    "requesting_user",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]
